#include <backtrace.h>

static void error_callback (void *data, const char *message, int error_number) {
  if (error_number == -1) {
    fprintf(stderr, "If you want backtraces, you have to compile with -g\n\n");
    _Exit(1);
  } else {
    fprintf(stderr, "Backtrace error %d: %s\n", error_number, message);
  };
};

static int full_callback (void *data, uintptr_t pc, const char *pathname, int line_number, const char *function) {
  static int unknown_count = 0;
  if (pathname != NULL || function != NULL || line_number != 0) {
    if (unknown_count) {
      fprintf(stderr, "    ...\n");
      unknown_count = 0;
    };
    const char *filename = rindex(pathname, '/');
    if (filename) filename++; else filename = pathname;
    log_printf (NULL, LERROR, "  %s:%d -- %", filename, line_number, function);
  } else {
    unknown_count++;
  };
  return 0;
};


struct backtrace_state *state;

void backtrace_sigsegv_handler (int number) {
  fprintf(stderr, "\r\n\e[1;31mAnd all the land was laid to burnination!\e[0m\n\n");
  backtrace_full(state, 0, full_callback, error_callback, NULL);
  
  //backtrace_print(state, 0, stderr);
  _Exit(1);
};