/*
LICENSE
-------
The Anti-Viral License

You may use this source code in any way you like provided that this source code
retains its original licensing terms.  Derivative works of this  source code
must also be licensed under this license,  however,  this  license does not
forbid compiling this source code alongside  source  code  of other licenses
provided that those licenses do not require a change in the licensing of this
source code.  Just use some comments to track what code is of what license,
like any intelligent and rational person might do.   After all, no sane person
thinks that if I glue two books together, I've made a derivative work and
require licensing to sell the result.  Were that the case, it would be illegal
to sell used books with notes written in the margins or papers inserted between
the pages.  Let's not be silly about this.  We all know what it means to derive
something from something else. 

...or, in simpler terms, this license is specifically incompatible with the GNU
GPL, but only because the GNU GPL requires that the GNU GPL be applied to all
source code which is part of a program that includes GNU GPL licensed source
code.  Many other free software licenses, such as the BSD license, make no such
demands and are therefore compatible with this license.   Only licenses which
require the relicensing of this source code are incompatible with this license.
Such licenses are viral in nature, and thus, this is the anti-viral license.
*/
#include "global.h"




struct d_menu_lookup_box {
	char info[40];
	char text[255];	
	int length;
	void *callback;
	int name;
	int value;
	struct user duser;
	int uid;
	char area[64][128];
	int num_areas;
	int area_num;

} d_menu_lookup_box;


static int build_area_list (struct user *puser, struct d_menu_lookup_box *d) {
	struct dirent *rdn;
	struct stat st;
	DIR *dn;
	char path[PATH_MAX];
	char temppath[PATH_MAX];
	int ret;
	char *cp;
	
	d->num_areas = 0;
	
	snprintf (path, PATH_MAX, "%s/%s/%i", PATH_AREAS, puser->map->file, puser->uid);

	log_printf (puser, LDEBUG, "area list path [%s]", path);

	dn = opendir (path);
    if (dn == 0) {
        log_printf (NULL, LERROR, "Unable to open directory [%s]: %s\n", strerror (errno), path);
		return (0);
    }
	
	while ((rdn = readdir (dn)) != NULL) {
		if (rdn->d_name[0] == '.') continue;	
		sprintf (temppath, "%.*s/%.256s", PATH_MAX - 257, path, rdn->d_name);
		
		ret = lstat (temppath, &st);
		if (ret == -1) {
            log_printf (NULL, LNOTE, "stat failed for [%s]: %s", temppath, strerror (errno));
            continue;
        }

		if (S_ISDIR (st.st_mode)) continue;
		
		_strncpy (d->area[d->num_areas], rdn->d_name, 128);
		
		cp = strrchr (d->area[d->num_areas], '.');
		if (cp == 0) continue;
		
		if (strcmp (cp, ".bin") != 0) continue;
		
		*cp = 0;
		
		d->num_areas++;
	}
	closedir (dn);
	return (1);
}


int menu_lookup_box (struct user *puser, int message_id, int name, int value, int area_num, char *username) {
	struct d_menu_lookup_box *d;
	
	d = menu_get_current_stack_item_memory (puser);

	switch (message_id) {
	case MENU_CREATE:
		log_printf (puser, LDEBUG, "menu_lookup_box");
		d = menu_stack_push_init (puser, menu_lookup_box, NULL, sizeof (struct d_menu_lookup_box));

		user_id_by_name (puser->map, (char *)username, &d->uid);

		if (!user_record_load (&d->duser, d->uid)) d->uid = 0;
		
		if (d->uid == 0) {
			menu_stack_pop (puser);	
			menu_dialog (puser, MENU_CREATE, "Load Area", "I don't know who that is!", "OK");			
			return (1);
		}
		
		d->area_num = area_num;
		d->name = name;
		d->value = value;

		// All of this happens on the same map.
		d->duser.map = puser->map;

		build_area_list (&d->duser, d);

		if (d->num_areas == 0) {
			menu_stack_pop (puser);	
			menu_dialog (puser, MENU_CREATE, "Load Area", "That user doesn't have any maps for some reason.", "OK");			
			return (1);
		}




	case MENU_DRAW: {
		int index = 1;


        packet_send(puser, PACKET_MENU_RESET, 0, 60, 4 + d->num_areas, 3); 

        packet_send(puser, PACKET_MENU_TEXT, index++, TEXT_CENTER_COLUMN, 0, "Select an area");
        for (int a = 0; a < d->num_areas; a++) {

        	packet_send(puser, PACKET_MENU_LINK, index++, 1, a + 2, 1, a + 10, 0, d->area[a]);
        }
        
        
		
		break;}
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		unsigned char menu_char[65535]; 
        menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, menu_char);

        if (menu_name == 1 && menu_value >= 10) {
        	menu_value -= 10;

			if (area_for_player (&d->duser) != -1) {
				menu_stack_pop (puser);
				char temp[512];
				sprintf (temp, "%s's area is already loaded.", d->duser.name);
				menu_dialog (puser, MENU_CREATE, "Load Area", temp, "OK");			
				return (1);
		
			}
		
			area_save (puser->map, d->area_num);
			memset (&puser->map->area[d->area_num].invited, 0, sizeof (size_t) * 256);
			
			area_record_clear (puser->map, d->area_num);
			
			puser->map->area[d->area_num].id = d->duser.uid;

			log_printf (puser, LDEBUG, "Selected area %d, %s", menu_value, d->area[menu_value]);
            	
            area_load (&d->duser, d->area_num, d->area[menu_value]);							


			

        }

		menu_stack_pop (puser);	
		

		break;}
	}

	
	return (1);		
}
