/*
LICENSE
-------
The Anti-Viral License

You may use this source code in any way you like provided that this source code
retains its original licensing terms.  Derivative works of this  source code
must also be licensed under this license,  however,  this  license does not
forbid compiling this source code alongside  source  code  of other licenses
provided that those licenses do not require a change in the licensing of this
source code.  Just use some comments to track what code is of what license,
like any intelligent and rational person might do.   After all, no sane person
thinks that if I glue two books together, I've made a derivative work and
require licensing to sell the result.  Were that the case, it would be illegal
to sell used books with notes written in the margins or papers inserted between
the pages.  Let's not be silly about this.  We all know what it means to derive
something from something else. 

...or, in simpler terms, this license is specifically incompatible with the GNU
GPL, but only because the GNU GPL requires that the GNU GPL be applied to all
source code which is part of a program that includes GNU GPL licensed source
code.  Many other free software licenses, such as the BSD license, make no such
demands and are therefore compatible with this license.   Only licenses which
require the relicensing of this source code are incompatible with this license.
Such licenses are viral in nature, and thus, this is the anti-viral license.
*/
#include "global.h"

struct d_menu_server_config {
	int id;
	struct config cfg;
	int max_rank;
	int mode;
	int current_rank;

	int (*menu_call)(struct user *puser, int message_id, int id);
};

#define ENTRY_TEXT(description, qrow, _name, _value, _flags, _text_length, _text_box_length, _text) ({\
		packet_send(puser, PACKET_MENU_TEXT, index++, 1, qrow, description);\
		packet_send(puser, PACKET_MENU_INPUT, index++, 30, qrow, _name, _value, _text_length, _text_box_length, _flags, _text);\
		})

#define ENTRY_TEXT_INT(description, _row, _name, _value, _flags, _text_length, _text_box_length, _text)({\
		sprintf (temp, "%d", _text);\
		packet_send(puser, PACKET_MENU_TEXT, index++, 1, _row, description);\
		packet_send(puser, PACKET_MENU_INPUT, index++, 30, _row, _name, _value, _text_length, _text_box_length, _flags, temp);\
		})

#define CHECK(description, _column, _row, _name, _value, _flags, _variable)({packet_send(puser, PACKET_MENU_CHECK, index++, 1, _row, _column, _flags, menu_set_check (_variable), description);})


static int sub_general (struct user *puser, int message_id, int id) {
	struct d_menu_server_config *d;
	d = menu_get_current_stack_item_memory (puser);

	char temp[1024];
	
	int index = 10;
	switch (message_id) {
	case MENU_DRAW:

		packet_send(puser, PACKET_MENU_TEXT, 1, TEXT_CENTER_COLUMN, 0, "Server Config - General");

		ENTRY_TEXT("Server Name", 4, 1, 200, 0, 16, 32, d->cfg.name);
		ENTRY_TEXT_INT("Connection Limit", 6, 2, 0, 0, 6, 6, d->cfg.connection_limit);
		ENTRY_TEXT_INT("Connection Timeout (seconds)", 8, 3, 0, 0, 6, 6, d->cfg.connection_timeout);
		ENTRY_TEXT_INT("AFK Timeout",10, 4, 0, 0, 6, 6, d->cfg.afk_timeout);
		ENTRY_TEXT("External Address", 12, 5, 0, 0, 16, 255, d->cfg.external_address);
		ENTRY_TEXT_INT("Restart Dead-Time", 14, 6, 0, 0, 6, 6, d->cfg.restart_dead_time);
		ENTRY_TEXT_INT("Minimum Client Version", 16, 7, 0, 0, 6, 6, d->cfg.minimum_client_version);

		CHECK("Allow Unauthenticated Connections", 1, 18, 8, 0, 0, d->cfg.local_connections);
		CHECK("Disable Server List Update", 1, 19, 9, 0, 0, d->cfg.announce_disabled);
		CHECK("Sync Configuration Changes", 1, 20, 10, 0, 0, d->cfg.config_sync);
		CHECK("Pause Server on Exit", 1, 21, 0, 0, 11, d->cfg.pause_exit);
		CHECK("Use ANSI Codes in Console", 1, 22, 12, 0, 0, d->cfg.ansi);
		CHECK("Print Log File Entries to the Console", 1, 23, 13, 0, 0, d->cfg.display_log);
		
		packet_send(puser, PACKET_MENU_TEXT, index++, 65, 4, "Default Rank:");
		
		if (rank_struct_promote_limits (puser->rank, &d->max_rank)) {
			int a, b;
			b = 5;
			for (a = 0; a < d->max_rank; a++) {
				if (rank[a].name[0] == 0) continue;
				int f = 0;
				if (strcasecmp (rank[a].name, d->cfg.default_rank_name) == 0) f = 1;
				packet_send(puser, PACKET_MENU_RADIO, index++, 66, b, 14, a, f, rank[a].name);
				b++;

			}
		}
		break;
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		char menu_char[65535]; 
		
		//int text_length = 
		menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, (unsigned char *)menu_char);
	
		if (menu_name == 1 && menu_value == 200) strncpy (d->cfg.name, menu_char, 32);
		else if (menu_name == 2 && menu_value == 0) d->cfg.connection_limit = atoi (menu_char);
		else if (menu_name == 3 && menu_value == 0) d->cfg.connection_timeout = atoi (menu_char);
		else if (menu_name == 4 && menu_value == 0) d->cfg.afk_timeout = atoi (menu_char);
		else if (menu_name == 5 && menu_value == 0) _strncpy (d->cfg.external_address, menu_char, 255);
		else if (menu_name == 6 && menu_value == 0) d->cfg.restart_dead_time = atoi (menu_char);
		else if (menu_name == 7 && menu_value == 0) d->cfg.minimum_client_version = atoi (menu_char);
		else if (menu_name == 8 && menu_value == 0) d->cfg.local_connections = menu_flags;
		else if (menu_name == 9 && menu_value == 0) d->cfg.announce_disabled = menu_flags;
		else if (menu_name == 10 && menu_value == 0) d->cfg.config_sync = menu_flags;
		else if (menu_name == 11 && menu_value == 0) d->cfg.pause_exit = menu_flags;
		else if (menu_name == 12 && menu_value == 0) d->cfg.ansi = menu_flags;
		else if (menu_name == 13 && menu_value == 0) d->cfg.display_log = menu_flags;
		else if (menu_name == 14 && (menu_value >= 0 && menu_value < d->max_rank)) strcpy (d->cfg.default_rank_name, rank[menu_value].name);
		else {
			log_printf (puser, LWARNING, "sub_general: unknown name=%i value=%i", menu_name, menu_value);
			menu_server_config (puser, MENU_DRAW, 0);
		}
	}
	}
	return (1);
}

static int sub_ranks (struct user *puser, int message_id, int id) {
	struct d_menu_server_config *d;
	d = menu_get_current_stack_item_memory (puser);

	char temp[1024];
	
	int index = 10;
	switch (message_id) {
	case MENU_DRAW:

		packet_send(puser, PACKET_MENU_TEXT, 1, TEXT_CENTER_COLUMN, 0, "Server Config - Ranks");

		packet_send(puser, PACKET_MENU_TEXT, index++, 1, 4, "Select Rank:");
		
		if (rank_struct_promote_limits (puser->rank, &d->max_rank)) {
			int a, b;
			b = 6;
			for (a = 0; a < d->max_rank; a++) {
				if (rank[a].name[0] == 0) continue;
				//char f = (1 << 1);
				//if (a == d->current_rank) f |= (1 << 0);
				//packet_send(puser, PACKET_MENU_RADIO, index++, 2, b, 2, a, f, rank[a].name);
				char temp[126];
				char *cc = " ";
				if (a == d->current_rank) cc = ">";
				sprintf (temp, "%s%s", cc, rank[a].name );
				//packet_send(puser, PACKET_MENU_LINK, index++, 2, b, 2, a, 0, temp);
				packet_send(puser, PACKET_MENU_BUTTON, index++, 0, b, 2, a, 16, (a == d->current_rank), rank[a].name);
				b += 2;
			}
		}

		packet_send(puser, PACKET_MENU_TEXT, index++, 20, 4, "Rank Name");
		packet_send(puser, PACKET_MENU_INPUT, index++, 44, 4, 3, 1, 16, 24, 0, rank[d->current_rank].name);

		packet_send(puser, PACKET_MENU_TEXT, index++, 20, 6, "MOTD file");
		packet_send(puser, PACKET_MENU_INPUT, index++, 44, 6, 3, 2, 16, 255, 0, rank[d->current_rank].motd_file);

/*		packet_send(puser, PACKET_MENU_TEXT, index++, 30, 8, "Allowed Commands");
		packet_send(puser, PACKET_MENU_INPUT, index++, 40, 8, 3, 3, 24, 255, 0, rank[d->current_rank].allowed_commands);

		packet_send(puser, PACKET_MENU_TEXT, index++, 30, 10, "Allowed Menus");
		packet_send(puser, PACKET_MENU_INPUT, index++, 40, 10, 3, 3, 24, 255, 0, rank[d->current_rank].allowed_menus);
*/
		packet_send(puser, PACKET_MENU_TEXT, index++, 20, 8, "Promote Max Rank");
		packet_send(puser, PACKET_MENU_INPUT, index++, 44, 8, 3, 3, 16, 255, 0, rank[d->current_rank].promote_to);

		sprintf (temp, "%g", rank[d->current_rank].cuboid_size_limit_meters);
		packet_send(puser, PACKET_MENU_TEXT, index++, 20, 10, "Cuboid Size (meters)");
		packet_send(puser, PACKET_MENU_INPUT, index++, 44, 10, 3, 3, 6, 6, 0, temp);

		sprintf (temp, "%g", rank[d->current_rank].cuboid_volume_limit_meters);
		packet_send(puser, PACKET_MENU_TEXT, index++, 20, 12, "Cuboid Volume (meters)");
		packet_send(puser, PACKET_MENU_INPUT, index++, 44, 12, 3, 3, 6, 6, 0, temp);

		packet_send(puser, PACKET_MENU_CHECK, index++, 65, 4, 5, 0, menu_set_check (rank[d->current_rank].can_chat), "Can Chat");
		packet_send(puser, PACKET_MENU_CHECK, index++, 65, 5, 5, 1, menu_set_check (rank[d->current_rank].can_build), "Can Build");
		packet_send(puser, PACKET_MENU_CHECK, index++, 65, 6, 5, 2, menu_set_check (rank[d->current_rank].can_fly), "Can Fly");
		packet_send(puser, PACKET_MENU_CHECK, index++, 65, 7, 5, 3, menu_set_check (rank[d->current_rank].can_sonic_fly), "Can Fly Fast");
		packet_send(puser, PACKET_MENU_CHECK, index++, 65, 8, 5, 4, menu_set_check (rank[d->current_rank].can_noclip), "Can Noclip");

		int c = 1;
		int r = 16;
		int idx = 0;

		while (!command_list_name_get (&idx, temp)) {
			int l = strlen (temp) + 5;
			if (c + l > puser->window_width) {
				c = 1;
				r++;
				if (r + 2 >= puser->window_height) {
					packet_send(puser, PACKET_MENU_TEXT, index++, 1, puser->window_height - 2, "Expand the game window to see all of the commands...");
					break;
				}
			} 
			packet_send(puser, PACKET_MENU_CHECK, index++, c, r, 6, idx, 0, temp);
			c += l;
			idx++;
		}


		break;
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		char menu_char[65535]; 
		
		//int text_length = 
		menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, (unsigned char *)menu_char);
	
		if (menu_name == 2) {
			d->current_rank = menu_value;
			printf ("2: %i\n", menu_value);
			menu_server_config (puser, MENU_DRAW, 0);
			return (1);
		} else {
			log_printf (puser, LWARNING, "sub_ranks: unknown name=%i value=%i", menu_name, menu_value);
			menu_server_config (puser, MENU_DRAW, 0);
		}
	}
	}
	return (1);
}

static int sub_maps (struct user *puser, int message_id, int id) {
	struct d_menu_server_config *d;
	d = menu_get_current_stack_item_memory (puser);

	char temp[1024];
	
	int index = 10;
	switch (message_id) {
	case MENU_DRAW:

		packet_send(puser, PACKET_MENU_TEXT, 1, TEXT_CENTER_COLUMN, 0, "Server Config - Maps");

		break;
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		char menu_char[65535]; 
		
		//int text_length = 
		menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, (unsigned char *)menu_char);
	
		if(0){}
		else {
			log_printf (puser, LWARNING, "sub_maps: unknown name=%i value=%i", menu_name, menu_value);
			menu_server_config (puser, MENU_DRAW, 0);
		}
	}
	}
	return (1);
}

static int sub_users (struct user *puser, int message_id, int id) {
	struct d_menu_server_config *d;
	d = menu_get_current_stack_item_memory (puser);

	char temp[1024];
	
	int index = 10;
	switch (message_id) {
	case MENU_DRAW:

		packet_send(puser, PACKET_MENU_TEXT, 1, TEXT_CENTER_COLUMN, 0, "Server Config - Users");

		break;
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		char menu_char[65535]; 
		
		//int text_length = 
		menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, (unsigned char *)menu_char);
	
		if(0){}
		else {
			log_printf (puser, LWARNING, "sub_users: unknown name=%i value=%i", menu_name, menu_value);
			menu_server_config (puser, MENU_DRAW, 0);
		}
	}
	}
	return (1);
}

int menu_server_config (struct user *puser, int message_id, int id) {
	struct d_menu_server_config *d;
	d = menu_get_current_stack_item_memory (puser);

	int index = 2;
	switch (message_id) {
	case MENU_CREATE:
		d = menu_stack_push_init (puser, menu_server_config, NULL, sizeof (struct d_menu_server_config));
		d->id = id;
		d->mode = 0;

		
		memmove (&d->cfg, &config, sizeof (config));
	case MENU_DRAW:
		if (d->mode == 1) packet_send(puser, PACKET_MENU_RESET, 0, puser->window_width - 1, puser->window_height - 1, 3); 
		else packet_send(puser, PACKET_MENU_RESET, 0, 80, 24, 3); 
		packet_send(puser, PACKET_MENU_BUTTON, index++, 0, 2, 253, 0, 8, (d->mode == 0), "General");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 10, 2, 253, 1, 8, (d->mode == 1), "Ranks");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 20, 2, 253, 2, 8, (d->mode == 2), "Maps");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 30, 2, 253, 3, 8, (d->mode == 3), "Users");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 70, puser->window_height - 3, 254, 254, 0, 0, "Apply");


		if (d->mode == 0) d->menu_call = sub_general;
		else if (d->mode == 1) d->menu_call = sub_ranks;
		else if (d->mode == 2) d->menu_call = sub_maps;
		else if (d->mode == 3) d->menu_call = sub_users;
		else {
			log_printf (puser, LERROR, "menu_server_config: bad menu mode %i\n", d->mode);
			return (1);
		}


		d->menu_call (puser, MENU_DRAW, id);

		break;
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		char menu_char[65535]; 
		
		menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, (unsigned char *)menu_char);
	
		if (menu_name == 253 && (menu_value >= 0 && menu_value < 4)) {
			d->mode = menu_value;
			menu_server_config (puser, MENU_DRAW, 0);
			return(1);

		} else if (menu_name == 254 && menu_value == 254) {

			char message[4096] = {0};
			if (strlen (d->cfg.name) < 4) strcat (message, "* Server name must be at least 4 characters.\n");
			if (d->cfg.connection_limit < 1 || d->cfg.connection_limit > 254) strcat (message, "* Connection limit must between 1 and 254\n");
			if (d->cfg.connection_timeout <= -1) strcat (message, "* Connection timeout must be 0 or greater.\n");
			if (d->cfg.afk_timeout <= -1) strcat (message, "* AFK timeout must be 0 or greater.\n");
			if (d->cfg.restart_dead_time <= -1) strcat (message, "* Restart Dead Time timeout must be 0 or greater.\n");
			if (d->cfg.minimum_client_version <= -1) strcat (message, "* Minimum Client Version must be 0 or greater.\n");

			if (message[0] != 0) {
				menu_dialog (puser, MENU_CREATE, "Server Configuration", message, "OK");			
				return (1);
			} else {
				memmove (&config, &d->cfg, sizeof (config));
				menu_stack_pop (puser);

				packet_broadcast (NULL, USER_CONNECTED, NULL, PACKET_SERVER_NAME, config.name);
				if (config.config_sync) config_save () ;

				return (1);
			}

		} else {
			if (menu_name == -1 && menu_value == -1) menu_stack_pop (puser);	
			else d->menu_call (puser, message_id, id);

			//printf ("number=%i name =%i, value=%i\n", menu_number, menu_name, menu_value);
		}
	}
	}
	return (1);		
	
}

/*int menu_server_config (struct user *puser, int message_id, int id) {
	struct d_menu_server_config *d;
	d = menu_get_current_stack_item_memory (puser);

	char temp[1024];
	
	int index = 1;
	switch (message_id) {
	case MENU_CREATE:
		d = menu_stack_push_init (puser, menu_server_config, NULL, sizeof (struct d_menu_server_config));
		d->id = id;		
		
		memmove (&d->cfg, &config, sizeof (config));
	case MENU_DRAW:
		packet_send(puser, PACKET_MENU_RESET, 200, 80, 24, 3); 
		packet_send(puser, PACKET_MENU_TEXT, index++, TEXT_CENTER_COLUMN, 0, "Server Config");

		#define ENTRY_TEXT(description, qrow, _name, _value, _flags, _text_length, _text_box_length, _text) ({\
				packet_send(puser, PACKET_MENU_TEXT, index++, 1, qrow, description);\
				packet_send(puser, PACKET_MENU_INPUT, index++, 30, qrow, _name, _value, _text_length, _text_box_length, _flags, _text);\
				})

		#define ENTRY_TEXT_INT(description, _row, _name, _value, _flags, _text_length, _text_box_length, _text)({\
				sprintf (temp, "%d", _text);\
				packet_send(puser, PACKET_MENU_TEXT, index++, 1, _row, description);\
				packet_send(puser, PACKET_MENU_INPUT, index++, 30, _row, _name, _value, _text_length, _text_box_length, _flags, temp);\
				})

		#define CHECK(description, _column, _row, _name, _value, _flags, _variable)({packet_send(puser, PACKET_MENU_CHECK, index++, 1, _row, _column, _flags, menu_set_check (_variable), description);})

		packet_send(puser, PACKET_MENU_BUTTON, index++, 0, 2, 253, 1, 0, 0, "General");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 8, 2, 253, 2, 0, 0, "Ranks");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 16, 2, 253, 3, 0, 0, "Maps");
		packet_send(puser, PACKET_MENU_BUTTON, index++, 24, 2, 253, 4, 0, 0, "Users");

		ENTRY_TEXT("Server Name", 4, 1, 200, 0, 16, 32, d->cfg.name);
		ENTRY_TEXT_INT("Connection Limit", 6, 2, 0, 0, 6, 6, d->cfg.connection_limit);
		ENTRY_TEXT_INT("Connection Timeout (seconds)", 8, 3, 0, 0, 6, 6, d->cfg.connection_timeout);
		ENTRY_TEXT_INT("AFK Timeout",10, 4, 0, 0, 6, 6, d->cfg.afk_timeout);
		ENTRY_TEXT("External Address", 12, 5, 0, 0, 16, 255, d->cfg.external_address);
		ENTRY_TEXT_INT("Restart Dead-Time", 14, 6, 0, 0, 6, 6, d->cfg.restart_dead_time);
		ENTRY_TEXT_INT("Minimum Client Version", 16, 7, 0, 0, 6, 6, d->cfg.minimum_client_version);

		CHECK("Allow Unauthenticated Connections", 1, 18, 8, 0, 0, d->cfg.local_connections);
		CHECK("Disable Server List Update", 1, 19, 9, 0, 0, d->cfg.announce_disabled);
		CHECK("Sync Configuration Changes", 1, 20, 10, 0, 0, d->cfg.config_sync);
		CHECK("Pause Server on Exit", 1, 21, 0, 0, 11, d->cfg.pause_exit);
		CHECK("Use ANSI Codes in Console", 1, 22, 12, 0, 0, d->cfg.ansi);
		CHECK("Print Log File Entries to the Console", 1, 23, 13, 0, 0, d->cfg.display_log);
		
		packet_send(puser, PACKET_MENU_TEXT, index++, 65, 4, "Default Rank:");
		
		if (rank_struct_promote_limits (puser->rank, &d->max_rank)) {
			int a, b;
			b = 5;
			for (a = 0; a < d->max_rank; a++) {
				if (rank[a].name[0] == 0) continue;
				int f = 0;
				if (strcasecmp (rank[a].name, d->cfg.default_rank_name) == 0) f = 1;
				packet_send(puser, PACKET_MENU_RADIO, index++, 66, b, 14, a, f, rank[a].name);
				b++;

			}
		}

		packet_send(puser, PACKET_MENU_BUTTON, index++, 70, 22, 254, 254, 0, 0, "Apply");		
	
		break;
		
	case MENU_COMMAND: {
		int menu_number = -1;
		int menu_name = -1;
		int menu_value = -1;
		int menu_flags = -1;
		char menu_char[65535]; 
		
		//int text_length = 
		menu_command_parse (puser, &menu_number, &menu_name, &menu_value, &menu_flags, (unsigned char *)menu_char);
		if (menu_number != 200) {
			printf ("Wrong menu: %i\n", menu_number);
			return (1);
		}
	
		if (menu_name == 1 && menu_value == 200) strncpy (d->cfg.name, menu_char, 32);
		else if (menu_name == 2 && menu_value == 0) d->cfg.connection_limit = atoi (menu_char);
		else if (menu_name == 3 && menu_value == 0) d->cfg.connection_timeout = atoi (menu_char);
		else if (menu_name == 4 && menu_value == 0) d->cfg.afk_timeout = atoi (menu_char);
		else if (menu_name == 5 && menu_value == 0) _strncpy (d->cfg.external_address, menu_char, 255);
		else if (menu_name == 6 && menu_value == 0) d->cfg.restart_dead_time = atoi (menu_char);
		else if (menu_name == 7 && menu_value == 0) d->cfg.minimum_client_version = atoi (menu_char);
		else if (menu_name == 8 && menu_value == 0) d->cfg.local_connections = menu_flags;
		else if (menu_name == 9 && menu_value == 0) d->cfg.announce_disabled = menu_flags;
		else if (menu_name == 10 && menu_value == 0) d->cfg.config_sync = menu_flags;
		else if (menu_name == 11 && menu_value == 0) d->cfg.pause_exit = menu_flags;
		else if (menu_name == 12 && menu_value == 0) d->cfg.ansi = menu_flags;
		else if (menu_name == 13 && menu_value == 0) d->cfg.display_log = menu_flags;
		else if (menu_name == 14 && (menu_value >= 0 && menu_value < d->max_rank)) strcpy (d->cfg.default_rank_name, rank[menu_value].name);

		else if (menu_name == 253 && menu_value == 2) {
			//menu_server_config_ranks (puser, MENU_CREATE, 0);
			menu_stack_set (puser, menu_server_config_ranks);
			menu_server_config_ranks (puser, MENU_DRAW, 0);
			return(1);

		} else if (menu_name == 254 && menu_value == 254) {

			char message[4096] = {0};
			if (strlen (d->cfg.name) < 4) strcat (message, "* Server name must be at least 4 characters.\n");
			if (d->cfg.connection_limit < 1 || d->cfg.connection_limit > 254) strcat (message, "* Connection limit must between 1 and 254\n");
			if (d->cfg.connection_timeout <= -1) strcat (message, "* Connection timeout must be 0 or greater.\n");
			if (d->cfg.afk_timeout <= -1) strcat (message, "* AFK timeout must be 0 or greater.\n");
			if (d->cfg.restart_dead_time <= -1) strcat (message, "* Restart Dead Time timeout must be 0 or greater.\n");
			if (d->cfg.minimum_client_version <= -1) strcat (message, "* Minimum Client Version must be 0 or greater.\n");

			if (message[0] != 0) {
				menu_dialog (puser, MENU_CREATE, "Server Configuration", message, "OK");			
				return (1);
			} else {
				memmove (&config, &d->cfg, sizeof (config));
				menu_stack_pop (puser);

				packet_broadcast (NULL, USER_CONNECTED, NULL, PACKET_SERVER_NAME, config.name);
				if (config.config_sync) config_save () ;

				return (1);
			}

		} else {
			if (menu_name == -1 && menu_value == -1) menu_stack_pop (puser);	
			//printf ("number=%i name =%i, value=%i\n", menu_number, menu_name, menu_value);
		}
	}
	}
	return (1);		
	
}*/
